/* adaptable generic build-tool driver (C-compile, link, librarian)
 * @(#) bld.c	1.1	27dec2016 MJS
 *
 * Copyright (c) 2016 Mike Spooner
 *----------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *----------------------------------------------------------------------
 * Core source-code conforms to ANSI standard X3.159-1989.
 * SDK-configuration files may use SDK-specific C extensions.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <assert.h>
#ifndef MAX_PATH
#define MAX_PATH 255 /* sic minima for win32, measured in bytes */
#endif

char *me = "?";
int noexec = 0;		/* only print the commands, do not execute them? */

#define C89  0
#define C89X 1
#define YES 1
#define NO 0

static int needquote(const char *str)
	{
	const char *p;

	assert(str);
	for (p = str; *p; ++p)
		{
		if (isspace(*p) || iscntrl(*p) || !isprint(*p) || (*p == '/'))
			return (YES);
		}
	return (NO);
	}

static void panic(char *msg)
	{
	fprintf(stderr, "%s panic: %s\n", me, msg);
	exit(EXIT_FAILURE);
	}

static void error(char *msg)
	{
	fprintf(stderr, "%s error: %s\n", me, msg);
	exit(EXIT_FAILURE);
	}

static void remark(char *msg)
	{
	fprintf(stderr, "%s: %s\n", me, msg);
	}

static void *xalloc(unsigned int size)
	{
	void *p;

	if (!size)
		panic("zero-length allocation");
	p = malloc(size);
	if (!p)
		panic("out of memory");
	return (p);
	}

static void *xrealloc(void *old, unsigned int size)
	{
	void *p;

	if (!size)
		panic("zero-length reallocation");
	if (!old)
		p = malloc(size);
	else
		p = realloc(old, size);
	if (!p)
		panic("out of memory");
	return (p);
	}

static void *xfree(void *old)
	{
	free(old);
	return (NULL);
	}

static char *str_buy(char *str, const char *suffix)
	{
	str = xrealloc(str, strlen(str) + strlen(suffix) + 1);
	strcat(str, suffix);
	return (str);
	}

static char *str_dup(const char *str)
	{
	char *p;

	p = xalloc(strlen(str) + 1);
	strcpy(p, str);
	return (p);
	}

#ifdef DEBUG
void vec_dump(char *name, const char **v)
	{
	if (v && name)
		{
		unsigned int i;
		fprintf(stderr, "%s: [%p]\n", name, v);
		for (i = 0; v[i]; ++i)
			{
			fprintf(stderr, "    [%i] at [%p] = ", i, v[i]);
			fprintf(stderr, "[%s]", v[i]);
			fprintf(stderr, "\n");
			}
		fprintf(stderr, "---------------\n");
		}
	}
#endif

/* print a string-vector as a sequence of space-separated strings:
 */
static void vec_print(char **v)
	{
	assert(v);

	while (*v)
		{
		printf("%s ", *v);
		++v;
		}
	printf("\n");
	}

/* create/extend a string-vector:
 */
static char **vec_buy(char **v, const char *str)
	{
	char *s;
	unsigned int i;

	if (!str)
		return (v);
	if (v)
		{
		for (i = 0; v[i]; ++i)
			;
		}
	else
		i = 0;
	v = xrealloc(v, (i + 2) * sizeof(*v));
	v[i++] = str_dup(str);
	v[i] = NULL;
	return (v);
	}

/* deallocate a string-vector and its' contents:
 */
static char **vec_discard(char **v)
	{
	if (v)
		{
		unsigned int i;
		for (i = 0; v[i]; ++i)
			xfree(v[i]);
		xfree(v);
		}
	return (NULL);
	}

static char *basename(const char *filename)
	{
	char *buf;
	const char *p;

	p = strrchr(filename, '.');
	if (p)
		{
		if (p == filename)
			return (str_dup(""));
		buf = xalloc((p - filename) + 1);
		strncpy(buf, filename, p - filename);
		buf[p - filename] = '\0';
		return (buf);
		}
	return (str_dup(filename));
	}

static char *ext(const char *filename, int lower)
	{
	char *buf;
	const char *p;

	p = strrchr(filename, '.');
	if (p)
		{
		if (*++p)
			{
			buf = xalloc(strlen(p) + 1);
			strcpy(buf, p);
			buf[strlen(p)] = '\0';
			if (lower)
				{
				char *q;
				for (q = buf; *q; ++q)
					*q = tolower(*q);
				}
			return (buf);
			}
		}
	return (str_dup(""));
	}

static int check_filename(const char *filename, ...)
	{
	va_list ap;
	char *b, *e;
	char *allowed;

	b = basename(filename);
	if (!strlen(b))
		{
		xfree(b);
		return (NO);
		}
	xfree(b);

	e = ext(filename, YES);
	va_start(ap, filename);
	while ((allowed = va_arg(ap, char *)) != NULL)
		{
		if (!strcmp(e, allowed))
			{
			xfree(e);
			return (YES);
			}
		}
	va_end(ap);
	xfree(e);
	return (NO);
	}

static void usage(void)
	{
	fprintf(stderr, "usage: %s %s\n",
		me, "c89 [ -i*^ -d*^ -n -o* -u*^ ] source-file");
	fprintf(stderr, "   or: %s %s\n",
		me, "c89x [ -i*^ -d*^ -n -o* -u*^ ] source-file");
	fprintf(stderr, "   or: %s %s\n",
		me, "link [ -l*^ -n -o* ] object-files [ library-files ]");
	fprintf(stderr, "   or: %s %s\n",
		me, "lib [ -n -o* ] object-files");
	exit(EXIT_FAILURE);
	}

static char *fq_pathname(const char *orig, char buf[MAX_PATH+5])
	{
	char *b, *e;

	b = basename(orig);
	e = ext(orig, NO);
	if (needquote(b) && needquote(e))
		sprintf(buf, "\"%s\".\"%s\"", b, e);
	else if (needquote(b))
		sprintf(buf, "\"%s\".%s", b, e);
	else if (needquote(e))
		sprintf(buf, "%s.\"%s\"", b, e);
	else
		strcpy(buf, orig);
	xfree(e);
	xfree(b);
	return (buf);
	}

static char *nq_pathname(const char *orig, char buf[MAX_PATH+5])
	{
	strcpy(buf, orig);
	return (buf);
	}

#include "bld.inc"

static void execute_spec(char *spec)
	{
	int status;
	char *p;
	char **cmd;

	cmd = NULL;
	for (p = strtok(spec, "|"); p; p = strtok(NULL, "|"))
		{
		if (strlen(p))
			cmd = vec_buy(cmd, p);
		}
	vec_print(cmd);
	if (!noexec)
		{
		/* NOTE1 runvp() *must* be a *synchronous* *child-process* call,
		 * ie: like UNIX fork+execvp+waitpid, or like Win32/DOS P_WAIT.
		 * 
		 * NOTE2 (void *) cast below is needed only to keep the daft
		 * Watcom conpiler quiet.
		 */
		status = runvp(*cmd, (void *) cmd);
		if (status)
			{
			fprintf(stderr,
				"error: failed to execute program: %s\n",
				cmd[0]);
			exit(EXIT_FAILURE);
			}
		}
	exit (EXIT_SUCCESS);
	}

static void opterror(const char *opt)
	{
	const char *pfx = "option ", *sfx = " expects an argument";
	char *buf;
	
	buf = xalloc(strlen(pfx)+strlen(opt)+2+1+strlen(sfx));
	sprintf(buf, "%s\"%s\"%s", pfx, opt, sfx);
	error(buf);
	}

static const char *getoval(int *ac, const char ***av, const char *opt)
	{
	int n;
	const char *arg;

	assert(*ac && *av && **av && ***av && opt && *opt);
	n = strlen(opt);
	if (!strncmp(**av, opt, n))
		{
		const char *p = &(**av)[n];
		if (!*p)	/* try following arg */
			{
			--(*ac), ++(*av);
			if (!**av)
				opterror(opt);
			arg = **av;
			}
		else
			arg = p;
		if (*p == '-')
			opterror(opt);
		--(*ac), ++(*av);
		return (arg);
		}
	return (NULL);
	}


/* replace first occurrence of sentinel with formatted tokens
 */
static char *spec_subst(char *spec, const char *sentinel, char **items,
	const char *fmt1, const char *fmt)
	{
	char *p, *tail, **v;
	const char *f, *q;
	int n;

	p = strstr(spec, sentinel);
	if (!p)	/* no placeholder present, presume not needed */
		return (spec);
	*p = '\0';
	tail = str_dup(p + strlen(sentinel));

	for (v = items; v && *v; ++v)
		{
		if (v == items) 	/* first one */
			f = fmt1;
		else
			f = fmt;
		assert(f);

		/* append the current (single) component
		 */
		n = strlen(f) + strlen(*v); /* slightly conservative */
		spec = xrealloc(spec, strlen(spec) + n + 1);
		for (p = spec + strlen(spec), q = f; q && *q; ++q)
			{
			if (*q == '\v')	/* placeholder found */
				{
				strcpy(p, *v);
				p += strlen(*v);
				}
			else
				*p++ = *q;
			}
		*p = '\0';
		}

	spec = str_buy(spec, tail);
	xfree(tail);
	return (spec);
	}

static void do_compile(int ac, const char **av, int mode)
	{
	const char *value;
	char *spec;
	char **dirs, **defs, **undefs, **infile, **outfile;

	infile = outfile = dirs = defs = undefs = NULL;

	while (ac && *av)
		{
		if (!strcmp(*av, "--"))	/* end of options */
			{
			--ac, ++av;
			break;
			}
		else if ((value = getoval(&ac, &av, "-i")) != 0)
			dirs = vec_buy(dirs, value);
		else if ((value = getoval(&ac, &av, "-d")) != 0)
			defs = vec_buy(defs, value);
		else if ((value = getoval(&ac, &av, "-u")) != 0)
			undefs = vec_buy(undefs, value);
		else if ((value = getoval(&ac, &av, "-o")) != 0)
			outfile = vec_buy(outfile, value);
		else if (!strcmp(*av, "-n"))
			{
			noexec = 1;
			--ac, ++av;
			}
		else if (**av == '-')	/* unrecognised option */
			usage();
		else
			break;
		}

	if (!outfile || !*outfile || !**outfile)
		error("missing -o* option");
	if (!check_filename(*outfile, OBJ_EXT, NULL))
		error("invalid output filename");

	if (ac && *av)
		{
		infile = vec_buy(infile, *av);
		--ac, ++av;
		}
	if (ac || *av || !infile || !*infile || !**infile || outfile[1])
		usage();
	if (check_filename(*infile, OBJ_EXT, EXE_EXT, SLL_EXT, NULL))
		error("invalid input filename");

	/* build command-line spec
	 */
	spec = str_dup((mode == C89) ? C89_SPEC : C89X_SPEC);
	spec = spec_subst(spec, "[predefs]", defs, CC_PREDEF1, CC_PREDEF);
	vec_discard(defs);
	spec = spec_subst(spec, "[undefs]", undefs, CC_UNDEF1, CC_UNDEF);
	vec_discard(undefs);
	spec = spec_subst(spec, "[incdirs]", dirs, CC_INCDIR1, CC_INCDIR);
	vec_discard(dirs);
	spec = spec_subst(spec, "[obj]", outfile, "\v", "\v");
	vec_discard(outfile);
	spec = spec_subst(spec, "[src]", infile, "\v", "\v");
	vec_discard(infile);

	execute_spec(spec);
	}

static void do_link(int ac, const char **av)
	{
	const char *value;
	char *spec, **dirs, **infiles, **outfile, **libs;

	dirs = infiles = outfile = libs = NULL;
	while (ac && *av)
		{
		if (!strcmp(*av, "--"))	/* explicit end of options */
			{
			--ac, ++av;
			break;
			}
		else if ((value = getoval(&ac, &av, "-l")) != 0)
			dirs = vec_buy(dirs, value);
		else if ((value = getoval(&ac, &av, "-o")) != 0)
			outfile = vec_buy(outfile, value);
		else if (!strcmp(*av, "-n"))
			{
			noexec = 1;
			--ac, ++av;
			}
		else if (**av == '-')	/* unrecognised option */
			usage();
		else
			break;
		}

	if (!outfile || !*outfile)
		error("missing -o* option");
	if (!check_filename(*outfile, EXE_EXT, NULL))
		error("invalid output filename");

	for (; ac && *av; --ac, ++av)
		{
		if (check_filename(*av, OBJ_EXT, NULL))	/* object-file */
			infiles = vec_buy(infiles, *av);
		else	/* not an object-file, presume it is a library-file */
			libs = vec_buy(libs, *av);
		}
	if (!infiles || !*infiles)
		usage();

	/* build command-line spec
	 */
	spec = str_dup(LNK_SPEC);
	/* libdirs */
	#if !defined(LNK_LIBDIR1) || !defined(LNK_LIBDIR)
	dirs = vec_buy(dirs, ""); /* also search for the original filename! */
	libs = findfiles((const char **) dirs, libs);
	spec = spec_subst(spec, "[libdirs]", dirs, "", "");
	#else
	spec = spec_subst(spec, "[libdirs]", dirs, LNK_LIBDIR1, LNK_LIBDIR);
	#endif
	vec_discard(dirs);
	/* library-files */
	spec = spec_subst(spec, "[libs]", libs, "\v|", "\v|");
	vec_discard(libs);
	/* object-files */
	spec = spec_subst(spec, "[objs]", infiles, "\v|", "\v|");
	vec_discard(infiles);
	/* output-file */
	spec = spec_subst(spec, "[exe]", outfile, "\v", "\v");
	vec_discard(outfile);

	execute_spec(spec);
	}

static void do_static_lib(int ac, const char **av)
	{
	const char *value;
	char *spec, **infiles, **outfile;

	infiles = outfile = NULL;
	while (ac && *av)
		{
		if (!strcmp(*av, "--"))
			{
			--ac, ++av;
			break;
			}
		else if ((value = getoval(&ac, &av, "-o")) != 0)
			outfile = vec_buy(outfile, value);
		else if (!strcmp(*av, "-n"))
			{
			noexec = 1;
			--ac, ++av;
			}
		else if (**av == '-')	/* unrecognised option */
			usage();
		else
			break;
		}

	if (!outfile || !*outfile)
		error("missing -o* option");
	if (!check_filename(*outfile, SLL_EXT, NULL))
		error("invalid output filename");

	for (; ac && *av; --ac, ++av)
		infiles = vec_buy(infiles, *av);
	if (!infiles || !*infiles)
		usage();

	/* build command-line spec
	 */
	spec = str_dup(SLL_SPEC);
	spec = spec_subst(spec, "[lib]", outfile, "\v", "\v");
	vec_discard(outfile);
	spec = spec_subst(spec, "[objs]", infiles, "\v|", "\v|");
	vec_discard(infiles);

	execute_spec(spec);
	}

int main(int ac, char **av)
	{
	void *np = NULL;

	setvbuf(stderr, np, _IONBF, 0);
	setvbuf(stdout, np, _IONBF, 0);

	if ((me = strrchr(*av, '\\')) != 0)
		++me;
	else if (isalpha(**av) && ((me = strchr(*av, ':')) != 0))
		++me;
	else
		me = *av;

	--ac, ++av;
	if (!ac || !av || !*av)
		usage();
	if (!strcmp(*av, "c89"))
		do_compile(--ac, (const char **) (++av), C89);
	else if (!strcmp(*av, "c89x"))
		do_compile(--ac, (const char **) (++av), C89X);
	else if (!strcmp(*av, "link"))
		do_link(--ac, (const char **) (++av));
	else if (!strcmp(*av, "lib"))
		do_static_lib(--ac, (const char **) (++av));

	usage();
	return (EXIT_FAILURE);
	}
